/*
 * SM501 USB HCD for Linux Version.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2003 (c) Lineo uSolutions,Inc.
 * Copyright 2004 (c) Paul Mundt
 */

#include <linux/dma-mapping.h>
#include <linux/init.h>
#include <linux/module.h>
#include <asm/rts7751r2d/voyagergx_reg.h>
#include <asm/bus-sh.h>

#ifndef CONFIG_VOYAGERGX
#error "This file is VoyagerGX glue.  CONFIG_VOYAGERGX must be defined."
#endif

extern int usb_disabled(void);

static void usb_hcd_voyager_remove(struct usb_hcd *hcd, struct sh_dev *dev);
static void hcd_voyager_release(struct usb_bus *bus);
static int ohci_voyager_reset(struct usb_hcd *hcd);
static int __devinit ohci_voyager_start(struct usb_hcd *hcd);
static int __init voyagergx_ohci_init(void);
static void __exit voyagergx_ohci_exit(void);

/*
 * RTS7751R2D has PCI bus, but VoyagerGX is connected SH Bus directoly
 */
#define VOYAGER_OHCI_NAME	"voyager-ohci"

static void __init voyagergx_ohci_configure(void)
{
	unsigned long           val;

        // Power Mode 0 Gate
        val = inl(POWER_MODE0_GATE);
        outl((val | POWER_MODE0_GATE_UH), POWER_MODE0_GATE);

	val = inl(POWER_MODE1_GATE);
        outl((val | POWER_MODE1_GATE_UH), POWER_MODE1_GATE);

        //Miscellaneous USB Clock Selsct
        val = inl(MISC_CTRL);
	val &= ~MISC_CTRL_USBCLK_48;
        outl(val, MISC_CTRL);

        // Interrupt Mask
        val = inl(VOYAGER_INT_MASK);
        val |= 0x00000040;
        outl(val, VOYAGER_INT_MASK);
}

static void hcd_voyager_release(struct usb_bus *bus)
{
	struct usb_hcd *hcd = bus->hcpriv;

	if (hcd)
		hcd->driver->hcd_free(hcd);
}

static int usb_hcd_voyager_probe(const struct hc_driver *driver,
				 struct usb_hcd **hcd_out,
				 struct sh_dev *dev)
{
	struct usb_hcd *hcd;
	int retval;

	if (!request_mem_region(dev->res.start,
				dev->res.end - dev->res.start + 1,
				hcd_name)) {
		dbg("controller already in use");
		return -EBUSY;
	}

	hcd = driver->hcd_alloc();
	if (hcd == NULL) {
		dbg("hcd alloc fail");
		retval = -ENOMEM;
		err("init Voyager USB fail, %d", retval);
		return retval;
	}

	hcd->regs = dev->mapbase;
	hcd->driver = (struct hc_driver *)driver;
	hcd->description = driver->description;
	hcd->irq = dev->irq[0];
	hcd->self.controller = &dev->dev;

	if ((retval = hcd_buffer_create(hcd)) != 0) {
		driver->hcd_free(hcd);
		err("init Voyager USB fail, %d", retval);
		return retval;
	}

	printk(KERN_INFO "ohci_hcd Voyager OHCI Host Controller\n");

	if (driver->reset && (retval = driver->reset(hcd)) < 0) {
		printk(KERN_ERR "ohci_hcd can't reset\n");
		driver->hcd_free(hcd);
		return retval;
	}
	hcd->state = USB_STATE_HALT;

	if ((retval=request_irq(hcd->irq, usb_hcd_irq, SA_SHIRQ,
					"usb-ohci", hcd))) {
		printk(KERN_ERR "ohci_hcd request interrupt %x fauled\n", hcd->irq);
		driver->hcd_free(hcd);
		err("init Voyager USB fail, %d", retval);
		return retval;
	}

	printk(KERN_INFO "ohci_hcd irq %x, io base %p\n", hcd->irq, dev->mapbase);

	usb_bus_init(&hcd->self);
	hcd->self.op = &usb_hcd_operations;
	hcd->self.hcpriv = (void *)hcd;
	hcd->product_desc = "SM501 OHCI";
	hcd->self.release = &hcd_voyager_release;
	hcd->self.bus_name = "shbus";

	INIT_LIST_HEAD(&hcd->dev_list);

	usb_register_bus(&hcd->self);

	if ((retval = driver->start(hcd)) < 0) {
		usb_hcd_voyager_remove(hcd, dev);
		return retval;
	}

	*hcd_out = hcd;
	return 0;
}

static void usb_hcd_voyager_remove(struct usb_hcd *hcd, struct sh_dev *dev)
{
	struct usb_device	*hub;

	info("ohci_hcd remove, state %x\n", hcd->state);

	if (in_interrupt())
		BUG();

	hub = hcd->self.root_hub;
	if (HCD_IS_RUNNING(hcd->state))
		hcd->state = USB_STATE_QUIESCING;

	dbg("ohci_hcd roothub graceful disconnect\n");
	usb_disconnect(&hub);

	hcd->driver->stop(hcd);
	hcd_buffer_destroy(hcd);
	hcd->state = USB_STATE_HALT;

	free_irq(hcd->irq, hcd);

	usb_deregister_bus(&hcd->self);
}

static int ohci_voyager_reset(struct usb_hcd *hcd)
{
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);

	ohci->regs = hcd->regs;
	return hc_reset(ohci);
}

static int __devinit ohci_voyager_start(struct usb_hcd *hcd)
{
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	int		ret;

	ohci->hcca = dma_alloc_coherent(hcd->self.controller,
			sizeof *ohci->hcca, &ohci->hcca_dma, 0);
	if (!ohci->hcca)
		return -ENOMEM;

	memset(ohci->hcca, 0, sizeof(struct ohci_hcca));
	if ((ret=ohci_mem_init(ohci)) < 0) {
		ohci_stop(hcd);
		return ret;
	}

	if (hc_start(ohci) < 0) {
		ohci_err(ohci, "can't start\n");
		ohci_stop(hcd);
		return -EBUSY;
	}

	create_debug_files(ohci);

	return 0;
}

static const struct hc_driver voyager_hc_driver = {
	.description		= hcd_name,

	/*
	 * generic hardare linkage
	 */
	.irq			= ohci_irq,
	.flags			= HCD_USB11,

	/*
	 * basic lifecycle operations
	 */
	.reset			= ohci_voyager_reset,
	.start			= ohci_voyager_start,
	.stop			= ohci_stop,

	/*
	 * memory lifecycle (except per-request)
	 */
	.hcd_alloc		= ohci_hcd_alloc,
	.hcd_free		= ohci_hcd_free,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue		= ohci_urb_enqueue,
	.urb_dequeue		= ohci_urb_dequeue,
	.endpoint_disable	= ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number	= ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data	= ohci_hub_status_data,
	.hub_control		= ohci_hub_control,
};

static int usb_hcd_voyager_drv_probe(struct sh_dev *dev)
{
	struct usb_hcd *hcd = NULL;
	int ret;

	if (usb_disabled())
		return -ENODEV;
	
	ret = usb_hcd_voyager_probe(&voyager_hc_driver, &hcd, dev);
	if (ret == 0)
		sh_set_drvdata(dev, hcd);
	
	return ret;
}

static int usb_hcd_voyager_drv_remove(struct sh_dev *dev)
{
	struct usb_hcd *hcd = sh_get_drvdata(dev);

	usb_hcd_voyager_remove(hcd, dev);
	sh_set_drvdata(dev, NULL);

	return 0;
}

/*
 * Driver definitions to register with SH Bus
 */
static struct sh_driver usb_hcd_voyager_driver = {
	.drv	= {
		.name	= VOYAGER_OHCI_NAME,
	},
	.dev_id		= SH_DEV_ID_USB_OHCI,
	.bus_id		= SH_BUS_VIRT,
	.probe		= usb_hcd_voyager_drv_probe,
	.remove		= usb_hcd_voyager_drv_remove,
};

static u64 voyager_dmamask = 0xffffffffUL;

/*
 * Device definitions to match the driver above
 */
static struct sh_dev usb_hcd_voyager_dev = {
	.name		= VOYAGER_OHCI_NAME,
	.dev_id		= SH_DEV_ID_USB_OHCI,
	.bus_id		= SH_BUS_VIRT,
	.mapbase	= (void *)VOYAGER_USBH_BASE,
	.dma_mask	= &voyager_dmamask,
	.res	= {
		.start	= VOYAGER_USBH_BASE,
		.end	= VOYAGER_USBH_BASE + 0xfff,
	},

	.irq	= {
		VOYAGER_USBH_IRQ,
	},
};

static int __init voyagergx_ohci_init(void)
{
	int ret;

	voyagergx_ohci_configure();

	ret = sh_driver_register(&usb_hcd_voyager_driver);
	if (ret != 0)
		return -ENODEV;
	
	ret = sh_device_register(&usb_hcd_voyager_dev);
	if (ret != 0) {
		sh_driver_unregister(&usb_hcd_voyager_driver);
		return -ENODEV;
	}

	return ret;
}

static void __exit voyagergx_ohci_exit(void)
{
	sh_device_unregister(&usb_hcd_voyager_dev);
	sh_driver_unregister(&usb_hcd_voyager_driver);
}

module_init(voyagergx_ohci_init);
module_exit(voyagergx_ohci_exit);
